#include <sstream>

#include <w3c/svg/SVGStylable.hpp>
#include <w3c/svg/SVGElement.hpp>
#include <w3c/svg/SVGRectElement.hpp>
#include <w3c/svg/SVGCircleElement.hpp>

#include <hsm/interaction/Protocol.h>
#include <hsm/svgl/SVGLBehaviour.h>

#include <hsm/svgl/SVGLWindow.h>
#include <hsm/svgl/Svgl.h>

#include <Protocol.h>


/* forward declarations *******************************************************/

class Rectangle;


/* RectHandle class declaration ***********************************************/

class RectHandle : public hsm::SVGLBehaviour, public hsm::Translate, public hsm::EnterLeave {
protected :
	svg::SVGStylable *handle;
	Rectangle *rect;
	
public :
	RectHandle(svg::SVGElement *e, hsm::SVGLWindow *w);
	
	virtual bool begin();
	virtual void end();
	
	virtual void enter();
	virtual void leave();	
};


/* Various Handle class declaration *******************************************/

class BrHandle : public RectHandle {
	static const hsm::SVGLBehaviour::Factory::Register< BrHandle > registration;
public :
	BrHandle(svg::SVGElement *e, hsm::SVGLWindow *w) : RectHandle(e, w) {}
	virtual void translate(const hsm::Point &p);
};

class BlHandle : public RectHandle {
	static const hsm::SVGLBehaviour::Factory::Register< BlHandle > registration;
public :
	BlHandle(svg::SVGElement *e, hsm::SVGLWindow *w) : RectHandle(e, w) {}
	virtual void translate(const hsm::Point &p);
};

class TrHandle : public RectHandle {
	static const hsm::SVGLBehaviour::Factory::Register< TrHandle > registration;
public :
	TrHandle(svg::SVGElement *e, hsm::SVGLWindow *w) : RectHandle(e, w) {}
	virtual void translate(const hsm::Point &p);
};

class TlHandle : public RectHandle {
	static const hsm::SVGLBehaviour::Factory::Register< TlHandle > registration;
public :
	TlHandle(svg::SVGElement *e, hsm::SVGLWindow *w) : RectHandle(e, w) {}
	virtual void translate(const hsm::Point &p);
};


/* Rectangle class declaration ************************************************/

class Rectangle : public hsm::SVGLBehaviour, public hsm::Translate, public RaiseLower, public Deletable, public Colorable {
	static const hsm::SVGLBehaviour::Factory::Register< Rectangle > registration;
	
	svg::SVGRectElement *rect;
	svg::SVGRectElement *brHandle;
	svg::SVGRectElement *blHandle;
	svg::SVGRectElement *trHandle;
	svg::SVGRectElement *tlHandle;

public :
	Rectangle(svg::SVGElement *e, hsm::SVGLWindow *w);
	virtual ~Rectangle();
	
	void resize(float dl, float dr, float dt, float db);
	
	virtual void translate(const hsm::Point &p);
	virtual void raise();
	virtual void lower();
	virtual void setColor(float r, float g, float b);
};



/* RectHandle implementation **************************************************/

RectHandle::RectHandle(svg::SVGElement *e, hsm::SVGLWindow *w) : hsm::SVGLBehaviour(e, w) {
	handle = dynamic_cast< svg::SVGStylable * >(elem);

	handle->setFill("black");
	handle->setOpacity(0.);

	hsm::svgl::getAttribute(elem, "hsm-arg:parent", rect);
}

bool RectHandle::begin() {
	return rect->begin();
}

void RectHandle::end() {
	rect->end();
}

void RectHandle::enter() {
	window->lock();
	handle->setOpacity(.5);
	window->unlock();
}

void RectHandle::leave() {
	window->lock();
	handle->setOpacity(0.);
	window->unlock();
}


/* Handles implementation *****************************************************/

void BrHandle::translate(const hsm::Point &p) {
	rect->resize(0, p/hsm::Point::X, 0, p/hsm::Point::Y);
}
const hsm::SVGLBehaviour::Factory::Register< BrHandle > BrHandle::registration("brHandle");

void BlHandle::translate(const hsm::Point &p) {
	rect->resize(p/hsm::Point::X, 0, 0, p/hsm::Point::Y);
}
const hsm::SVGLBehaviour::Factory::Register< BlHandle > BlHandle::registration("blHandle");

void TrHandle::translate(const hsm::Point &p) {
	rect->resize(0, p/hsm::Point::X, p/hsm::Point::Y, 0);
}
const hsm::SVGLBehaviour::Factory::Register< TrHandle > TrHandle::registration("trHandle");

void TlHandle::translate(const hsm::Point &p) {
	rect->resize(p/hsm::Point::X, 0, p/hsm::Point::Y, 0);
}
const hsm::SVGLBehaviour::Factory::Register< TlHandle > TlHandle::registration("tlHandle");


/* Rectangle implementation ***************************************************/

Rectangle::Rectangle(svg::SVGElement *e, hsm::SVGLWindow *w) : hsm::SVGLBehaviour(e, w) {
	// main object 
	rect = dynamic_cast< svg::SVGRectElement * >(elem); // TODO : check if rect != 0
	
	// handles
	float X = rect->getX().getValue();
	float Y = rect->getY().getValue();
	float W = rect->getWidth().getValue();
	float H = rect->getHeight().getValue();

	brHandle = new svg::SVGRectElement(window->svgDocument);
	brHandle->setX(X + W - 5);
	brHandle->setY(Y + H - 5);
	brHandle->setHeight(9);
	brHandle->setWidth(9);
	window->svgFeedback->appendChild(brHandle);
	hsm::svgl::setAttribute(brHandle, "hsm:behaviour", "brHandle");
	hsm::svgl::setAttribute(brHandle, "hsm-arg:parent", this);
	window->instanciateBehaviour(brHandle);

	blHandle = new svg::SVGRectElement(window->svgDocument);
	blHandle->setX(X - 5);
	blHandle->setY(Y + H - 5);
	blHandle->setHeight(9);
	blHandle->setWidth(9);
	window->svgFeedback->appendChild(blHandle);
	hsm::svgl::setAttribute(blHandle, "hsm:behaviour", "blHandle");
	hsm::svgl::setAttribute(blHandle, "hsm-arg:parent", this);
	window->instanciateBehaviour(blHandle);

	trHandle = new svg::SVGRectElement(window->svgDocument);
	trHandle->setX(X + W - 5);
	trHandle->setY(Y - 5);
	trHandle->setHeight(9);
	trHandle->setWidth(9);
	window->svgFeedback->appendChild(trHandle);
	hsm::svgl::setAttribute(trHandle, "hsm:behaviour", "trHandle");
	hsm::svgl::setAttribute(trHandle, "hsm-arg:parent", this);
	window->instanciateBehaviour(trHandle);
	
	tlHandle = new svg::SVGRectElement(window->svgDocument);
	tlHandle->setX(X - 5);
	tlHandle->setY(Y - 5);
	tlHandle->setHeight(9);
	tlHandle->setWidth(9);
	window->svgFeedback->appendChild(tlHandle);
	hsm::svgl::setAttribute(tlHandle, "hsm:behaviour", "tlHandle");
	hsm::svgl::setAttribute(tlHandle, "hsm-arg:parent", this);
	window->instanciateBehaviour(tlHandle);
}

Rectangle::~Rectangle() {
	delete hsm::SVGLBehaviour::Factory::get(brHandle);
	delete hsm::SVGLBehaviour::Factory::get(blHandle);
	delete hsm::SVGLBehaviour::Factory::get(trHandle);
	delete hsm::SVGLBehaviour::Factory::get(tlHandle);
	window->lock();
	elem->getParentNode()->removeChild(elem);
	window->svgFeedback->removeChild(brHandle);
	window->svgFeedback->removeChild(blHandle);
	window->svgFeedback->removeChild(trHandle);
	window->svgFeedback->removeChild(tlHandle);
	window->unlock();
}


void Rectangle::resize(float dl, float dr, float dt, float db) {
	window->lock();
	float X = rect->getX().getValue() + dl;
	float Y = rect->getY().getValue() + dt;
	float W = rect->getWidth().getValue() - dl + dr;
	float H = rect->getHeight().getValue() - dt + db;

	rect->setX(X);
	rect->setWidth(W);
	rect->setY(Y);
	rect->setHeight(H);

	brHandle->setX(X + W - 5);
	brHandle->setY(Y + H - 5);

	blHandle->setX(X - 5);
	blHandle->setY(Y + H - 5);

	trHandle->setX(X + W - 5);
	trHandle->setY(Y - 5);

	tlHandle->setX(X - 5);
	tlHandle->setY(Y - 5);

	window->unlock();
}

void Rectangle::translate(const hsm::Point &p) {
	resize(p/hsm::Point::X, p/hsm::Point::X, p/hsm::Point::Y, p/hsm::Point::Y);
}

void Rectangle::raise() {
	rect->raise();
	blHandle->raise();
	trHandle->raise();
	tlHandle->raise();
	brHandle->raise();
}

void Rectangle::lower() {
	rect->lower();
}

void Rectangle::setColor(float r, float g, float b) {
	std::stringstream color;
	color << "rgb(" << int(r) << "," << int(g) << "," << int(b) << ")";
	rect->setFill(color.str().c_str());
}


const hsm::SVGLBehaviour::Factory::Register< Rectangle > Rectangle::registration("rectangle");
