////////////////////////////////////////////////////////////////////////////////
// File    : Component.h
// Content : a common interface class for hsm component
// History : 16-Oct-2001 - [rb] creation
//           11-Dec-2001 - [rb] added serialization
//           16-Jan-2002 - [rb] added externalisation of event processing
//                              strategy into EventProcessor object
//           19-Feb-2003 - [rb] added FunctorComponent
////////////////////////////////////////////////////////////////////////////////

/*
    Copyright (C) 2004  Renaud Blanch <rb.hsm@laposte.net>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


/// \file Component.h
/// libhsm components definition.


#if !defined(HSM_COMPONENT_H)
#define HSM_COMPONENT_H


/* includes *******************************************************************/

#include <hsm/core/Processing.h>

#include <set>
#include <string>


namespace hsm {
	
/* Component class declaration ************************************************/

	/// component class definition
	class Component {
	private :
		static unsigned int cn;

	public :
		typedef std::string Id; ///< component id type

	protected :
		typedef std::set< Component * > ComponentSet;

	protected :
		Id componentId;

	public :
		/// constructor
		/// \param id id of the component
		Component(const Id &id = Id());
		virtual ~Component(); ///< destructor

		/// gets the component id
		/// \return the component id
		Id getId() const;


	// emitter
	protected :
		ComponentSet receivers;

	public :
		/// adds a receiver
		/// \param receiver the receiver to add
		void addReceiver(Component *receiver);

		/// removes a receiver
		/// \param receiver the receiver to remove
		void removeReceiver(Component *receiver);

	protected :
		void dispatch(const Event &event);
		void dispatch(Event::Message m = Event::CHANGE, Data d = 0);


	// receiver
	protected :
		ComponentSet emitters;

	public :
		/// posts an event to the component
		/// \param event the event to post
		void post(const Event &event);

		/// posts an event to the component
		/// \param m message to post
		/// \param d data to post
		void post(Event::Message m = Event::CHANGE, Data d = 0);

	protected :
		void addEmitter(Component *emmiter);
		void removeEmitter(Component *emitter);

		virtual void process(const Event &event) {}
		virtual void doPost(const Event &event) { process(event); }
	};
	
	/// casts a component reference into a component subclass reference
	template< class C > C &operator/(Component &component, const C *) {
		COMPILE_ASSERT(BASE_OF(Component, C), Trying_to_cast_component_into_something);
		C *c = dynamic_cast< C * >(&component);
		return *c;
	}


	/// implementations
	namespace impl {
	
/* impl::Component class definition *******************************************/

		/// component class definition implementation
		template< class Base = hsm::Component, template< class > class Synchro = processing::Sync, int Fn = processing::Callback >
		class Component;

		/// specialisation for event processing using the normal hsm::Component::process
		/// method (which should be specialized to do something).
		template< class Base, template< class > class Synchro >
		class Component< Base, Synchro, processing::Callback > : public Synchro< Base > {
		public :
			/// constructor
			/// \param id the component id
			Component(const hsm::Component::Id &id = hsm::Component::Id()) : Synchro< Base >(id) {}
		};

		/// specialisation for event processing using a hsm::Functor given as a constructor
		/// argument.
		template< class Base, template< class > class Synchro >
		class Component< Base, Synchro, processing::Functor > : public Synchro< Base > {
		private :
			Functor *fn;

		public :
			/// constructor
			/// \param f the functor that will handle the event processing
			/// \param id the component id
			Component(Functor *f, const hsm::Component::Id &id = hsm::Component::Id()) : Synchro< Base >(id), fn(f) {}

		protected :
			virtual void process(const Event &event) { (*fn)(event.data); }
		};


	} // namespace impl


	
/* forward declaration ********************************************************/
	
	class Device;

	namespace repository {

/* Component repository *******************************************************/
		
		/// gets a component
		/// searchs the component in the repository and try to load it if it is
		/// not present.
		/// \return a pointer on the requested component
		/// \param id the name of the requested component
		Component *getComponent(const Component::Id &id);
		
		/// gets an system device event handler
		/// searchs the component in the repository and try to load it if it is
		/// not present. trys to dynamic_cast it into hsm::EvenHandler * .
		/// \return a pointer on the requested device
		/// \param id the name of the requested device
		Device *getDevice(const Component::Id &id);
		
		
		/// adds a component to the repository
		/// \param component a pointer on the component to add
		void add(Component *component);
		
		/// removes a component from the repository
		/// \param component a pointer on the component to remove
		void remove(Component *component);

		
	} // namespace repository
	
} // namespace hsm


#endif // !defined(HSM_COMPONENT_H)
