////////////////////////////////////////////////////////////////////////////////
// File    : Protocol.h
// Content : interaction protocols
// History : 25-Aug-2004 - [rb] creation
////////////////////////////////////////////////////////////////////////////////

/*
    Copyright (C) 2004  Renaud Blanch <rb.hsm@laposte.net>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


/// \file Protocol.h
/// Basic abstract interaction protocols definitions


#if !defined(HSM_PROTOCOL_H)
#define HSM_PROTOCOL_H


/* includes *******************************************************************/

#include <hsm/component/Point.h>

	
namespace hsm {

/* Protocol struct definition *************************************************/

	/// Common abstract protocol interface
	struct Protocol {
	protected :
		Point pick; ///< where the object has been picked
		
	public :
	
		Protocol() : pick(2) {} ///< constructor
		virtual ~Protocol()  {} ///< destructor

		/// returns whether the caller can start interacting or not
		/// \return yes or no
		virtual bool begin() { return true; }
		virtual void end()   {} ///< ending an interaction

		/// set where the object has been picked
		void setPickPoint(const Point &p) { pick = p; }
	};
	

/* EnterLeave protocol declaration ********************************************/

	/// Things that can be entered and leaved
	struct EnterLeave : virtual Protocol {
		virtual void enter() {} ///< what to do when entered
		virtual void leave() {} ///< what to do when lived 
	};
	
	/// Code wrapper around EnterLeave protocol
	template< class Implementation >
	struct EnterLeaveMixin : EnterLeave, virtual Implementation {
		virtual void enter() { Implementation::enter(); }
		virtual void leave() { Implementation::leave(); }
	};
	
	
/* PressRelease protocol declaration ******************************************/

	/// Things that can be pressed and released
	struct PressRelease : virtual Protocol {
		virtual void press()   {} ///< what if pressed ?
		virtual void release() {} ///< and what if released ?
	};
	
	/// Code wrapper around PressRelease protocol
	template< class Implementation >
	struct PressMixin : virtual PressRelease, virtual Implementation {
		virtual void press	() { Implementation::press(); }
	};

	/// Code wrapper around PressRelease protocol
	template< class Implementation >
	struct ReleaseMixin : virtual PressRelease, virtual Implementation {
		virtual void release()  { Implementation::release(); }
	};

	/// Code wrapper around PressRelease protocol
	template< class Implementation >
	struct PressReleaseMixin : PressMixin< Implementation >, ReleaseMixin< Implementation > {
	};
	
	
/* SelectDeselect protocol declaration ****************************************/

	/// Thing that can be (de)selected
	struct Select : virtual Protocol {
		/// what to do when (de)selected
		/// \param s selected or deselected ?
		virtual void select()   {}
		virtual void deselect() {}
	};
	
	/// Code wrapper around Select protocol
	template< class Implementation >
	struct SelectMixin : Select, virtual Implementation {
		virtual void select()   { Implementation::select(); }
		virtual void deselect() { Implementation::deselect(); }
	};

	
/* Translate protocol declaration *********************************************/

	/// Things that can be Translated
	struct Translate : virtual Protocol {
		/// what to do when translated
		/// \param dp delta dp
		virtual void translate(const Point &dp) {}
	};
	
	/// Code wrapper around Translate protocol
	template< class Implementation >
	struct TranslateMixin : Translate, virtual Implementation {
		virtual void translate(const Point &dp) { Implementation::translate(dp); }
	};


/* Rotate protocol declaration ************************************************/

	/// Things that can be Rotated
	struct Rotate : virtual Protocol {
		/// what to do when translated
		/// \param da delta angle
		virtual void rotate(float da = 0.) {}
	};
	
	/// Code wrapper around Rotate protocol
	template< class Implementation >
	struct RotateMixin : Rotate, virtual Implementation {
		virtual void rotate(float da = 0.) { Implementation::rotate(da); }
	};
	
	
/* Zoom protocol declaration **************************************************/

	struct Zoom : virtual Protocol {
		/// what to do when scaled
		/// \param ds delta scale (1. is neutral, ds is multiplicative ie. ds = 2. doubles, ds = .5 is the inverse)
		virtual void scale(float ds = 1.) {}

		/// what to do when zoomed
		/// \param dz delta zoom (0. is neutral, dz is additive, dz == 1. => +1%)
		inline void zoom(float dz = 0.) { scale(exp(dz * 0.01)); }
	};
	
	/// Code wrapper around Zoom protocol
	template< class Implementation >
	struct ZoomMixin : Zoom, virtual Implementation {
		virtual void scale(float ds = 1.) { Implementation::scale(ds); }
	};
	

} // namespace hsm


#endif // !defined(HSM_PROTOCOL_H)
