////////////////////////////////////////////////////////////////////////////////
// File    : DragAndDrop.cpp
// Content : drag and drop implementation
// History : 05-AOU-2005 - [rb] creation
////////////////////////////////////////////////////////////////////////////////


/* includes *******************************************************************/

#include "DragAndDrop.h"

#include <string>
#include <iostream>
using namespace std;

#include <hsm/svgl/Svgl.h>
#include <hsm/system/thread/Mutex.h>
#include <hsm/system/thread/ThreadLoop.h>
#include <hsm/util/Functor.h>
#include <hsm/svgl/SVGLWindow.h>
#include <hsm/svgl/SVGLWindowPicking.h>

#include <w3c/svg/SVGTransforms.hpp>
#include <w3c/svg/SVGTransformable.hpp>
#include <w3c/svg/SVGUseElement.hpp>


namespace hsm {

/* Draggable implementation ***************************************************/

	Draggable::Draggable(svg::SVGElement *e, SVGLWindow *w) : 
		SVGLBehaviour(e, w),
		dropZone(0),
		origin(2),
		position(2),
		replacer(0) {

		ghost = new svg::SVGUseElement(window->svgDocument);
		ghost->setFill("gray");

		svg::SVGTransformable *transformable = dynamic_cast< svg::SVGTransformable * >(ghost);
		trans = transformable->getTranslate(true);
	
		singleInteraction = new Mutex();
	}

	Draggable::~Draggable() {
		singleInteraction->lock();
		if(replacer != 0) {
			delete replacer;
		}
		delete singleInteraction;
	}

	bool Draggable::begin() {
		if(not singleInteraction->trylock()) {
			// one by one please !
			return false;
		}

		window->lock();
		{
			if(replacer != 0) {
				// was replacing from previous interaction
				delete replacer;
				replacer = 0;
			}
		
			// add the ghost
			window->window2screen(pick, origin);
			position = origin;

			setupGhost();
			window->svgCursors->appendChild(ghost);
		}
		window->unlock();
	
		return true;
	}

	void Draggable::end() {
		if(dropZone != 0) {
			// to be trashed
			window->lock();
			window->svgCursors->removeChild(ghost);
			window->unlock();

			dropZone->leave();
		
			// delegates drop action
			dropZone->drop(this);
		
		} else {
			// to be replaced
			interruptedReplace = true;
			replacer = new ThreadLoop(Functor::create(this, &Draggable::stepReplace), 50, 0, 0,
			                          Functor::create(this, &Draggable::removeGhost));
			replacer->play();
		}
	
		singleInteraction->unlock();
	}


	void Draggable::translate(const Point &p) {
		// translate ...
		window->lock();
		trans->tx += p/Point::X;
		trans->ty += p/Point::Y;
		window->unlock();

		// ... and look for the drop zone
		position += p;
		DropZone *dz = SVGLBehaviour::window->pick< DropZone >(position, 0, false);
		dz = (dz != 0 ? dz->checkAccept(this) : 0);
		if(dz != dropZone) {
			if(dropZone != 0) dropZone->leave();
			dropZone = dz;
			if(dropZone != 0) dropZone->enter();
		}
	}

	void Draggable::setupGhost() {
		string id;
		svgl::getAttribute(elem, "id", id);		
	
		string href = "#"; href += id;
		ghost->setHref(href.c_str());
	
		float sx, sy, tx, ty;
		svgl::getGlobalTransform(elem, sx, sy, tx, ty);
		trans->tx = tx;
		trans->ty = ty;
	}

	void Draggable::stepReplace() {
		//  replacing to original position
		static const float stepSize = 20.;
	
		Point delta(2);
		delta  = position;
		delta -= origin;
		float n = abs(delta);

		window->lock();
		{
			if(n <= stepSize) {
				// close enough, remove & stop
				window->svgCursors->removeChild(ghost);
				interruptedReplace = false;
			
				ThreadLoop::get()->stop();

			} else {
				// gets closer
				delta *= stepSize / n;
				position -= delta;

				trans->tx -= delta/Point::X;
				trans->ty -= delta/Point::Y;
			}
		}
		window->unlock();
	}

	void Draggable::removeGhost() {
		if(interruptedReplace) {
			window->svgCursors->removeChild(ghost);
		}
	}

} // namespace hsm
