////////////////////////////////////////////////////////////////////////////////
// File    : GLWindow.MACOSX.cpp
// Content : GLWindow::os struct implementation
// History : 11-Oct-2001 - [rb] creation
////////////////////////////////////////////////////////////////////////////////

/*
    Copyright (C) 2004  Renaud Blanch <rb.hsm@laposte.net>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


/* includes *******************************************************************/

#include <GLWindow.h>

#include <cstdlib> // getenv
#include <string>
#include <sstream>

#include <hsm/component/Point.h>
#include <hsm/util/Debug.h>

/* helper function ************************************************************/

namespace helper {
	static AGLPixelFormat getPixelFormat(bool full = false);
}	


namespace hsm {

/* GLWindow::os struct implementation *****************************************/

	void GLWindow::os::createGlContext(GLWindow *glw) {
		SetPortWindowPort(glw->window);

		// creating the openGl context
		AGLPixelFormat apf = helper::getPixelFormat();
		glw->glWindowContext = aglCreateContext(apf, 0);
		aglDestroyPixelFormat(apf);
		
		glw->glFullScreenContext = 0;

		glw->glCurrentContext = glw->glWindowContext;
	}

	void GLWindow::os::destroyGlContext(GLWindow *glw) {
		setFullScreen(glw, false);
		
		aglSetCurrentContext(0);
		aglSetDrawable(glw->glCurrentContext, 0);
		aglDestroyContext(glw->glCurrentContext);		
	}
	
	void GLWindow::os::makeCurrent(GLWindow *glw) {
		if(glw->glCurrentContext == glw->glWindowContext) 
			aglSetDrawable(glw->glCurrentContext, GetWindowPort(glw->window));
		aglSetCurrentContext(glw->glCurrentContext);
	}
	
	void GLWindow::os::swapBuffers(GLWindow *glw) {
		aglSwapBuffers(glw->glCurrentContext);	
	}

	void GLWindow::os::setVBLSync(GLWindow *glw, long vbl /* = 1 */) {
		aglSetInteger(glw->glCurrentContext, AGL_SWAP_INTERVAL, &vbl);
	}

	void GLWindow::os::setFullScreen(GLWindow *glw, bool full /* = true */) {
		// code adapted from nicolas roussel's nucleo
		// [http://insitu.lri.fr/~roussel/projects/nucleo/]
		
		if(full) {
			// switching to full screen

			if(glw->glFullScreenContext != 0) 
				return;

			SetUserFocusWindow(glw->window);
			CGCaptureAllDisplays();

			// copying full screen context from window context
			AGLPixelFormat apf = helper::getPixelFormat(true);
			glw->glFullScreenContext = aglCreateContext(apf, glw->glWindowContext);
			
			aglCopyContext(glw->glWindowContext, glw->glFullScreenContext, GL_ALL_ATTRIB_BITS);
			
			GLint swap_interval = 0;
			aglGetInteger(glw->glWindowContext, AGL_SWAP_INTERVAL, &swap_interval);
			aglSetInteger(glw->glFullScreenContext, AGL_SWAP_INTERVAL, &swap_interval);
			
			// setting full screen context as current context
			aglSetCurrentContext(glw->glFullScreenContext);
			aglUpdateContext(glw->glFullScreenContext);
			aglSetFullScreen(glw->glFullScreenContext, 0, 0, 0, 0);
			glw->glCurrentContext = glw->glFullScreenContext;

			// getting size and position
			GLint params[3];
			aglGetInteger(glw->glFullScreenContext, AGL_FULLSCREEN, params);
			*(glw->position) = 0, 0;
			*(glw->size) = params[0], params[1];
			
//			glw->doInit();
//			glw->reshape(params[0], params[1]);
			
		} else {
			// switching to window
			if(glw->glFullScreenContext == 0)
				return;
				
			aglSetDrawable(glw->glFullScreenContext, NULL);
			aglCopyContext(glw->glFullScreenContext, glw->glWindowContext, GL_ALL_ATTRIB_BITS);
			aglDestroyContext(glw->glFullScreenContext);

			glw->glCurrentContext = glw->glWindowContext;
			aglSetCurrentContext(glw->glCurrentContext);
			
			glw->glFullScreenContext = 0;

			CGReleaseAllDisplays();

			Rect rect ;
			GetWindowBounds(glw->window, kWindowContentRgn, &rect);
			*(glw->position) = rect.left, rect.top;
			*(glw->size) = rect.right - rect.left, rect.bottom - rect.top;
		}
	}

	void GLWindow::os::reshape(GLWindow *glw) {
		aglUpdateContext(glw->glCurrentContext);	
	}


} // namespace hsm


namespace helper {

/* helper function ************************************************************/

	static AGLPixelFormat getPixelFormat(bool full) {
		unsigned int next = 0;
		GLint attrib[] = { AGL_NONE, AGL_NONE, AGL_NONE, AGL_NONE,
		                   AGL_NONE, AGL_NONE, AGL_NONE, AGL_NONE,
		                   AGL_NONE, AGL_NONE, AGL_NONE, AGL_NONE,
		                   AGL_NONE, AGL_NONE, AGL_NONE, AGL_NONE };
		
		// fullscreen ?
		if(full)
			attrib[next++] = AGL_FULLSCREEN;
			
		// default setup
		attrib[next++] = AGL_MP_SAFE;
		attrib[next++] = AGL_NO_RECOVERY;		
		attrib[next++] = AGL_RGBA;
		attrib[next++] = AGL_DOUBLEBUFFER;
		attrib[next++] = AGL_PIXEL_SIZE;   attrib[next++] = 32;
		attrib[next++] = AGL_STENCIL_SIZE; attrib[next++] = 8;
		attrib[next++] = AGL_DEPTH_SIZE;   attrib[next++] = 32;
		
		// test for multisample
		char *cgs = getenv("HSM_GL_SAMPLES");
		unsigned int gs = 0;
		if(cgs != 0) {
			std::string sgs = cgs;
			std::stringstream strgs; strgs << sgs;
			strgs >> gs;
			attrib[next++] = AGL_SAMPLE_BUFFERS_ARB; attrib[next++] = 1;
			attrib[next++] = AGL_SAMPLES_ARB;        attrib[next++] = gs;
		}
		
		GDHandle device = GetMainDevice();
		return aglChoosePixelFormat(&device, 1, attrib);	
	}
	
} // namespace
